/*
 * Copyright 2005-2007 WSO2, Inc. (http://wso2.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 * http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.solutions.identity.relyingparty.service;

import java.io.ByteArrayInputStream;
import java.io.FileInputStream;
import java.io.StringReader;
import java.net.URI;
import java.security.KeyStore;
import java.security.MessageDigest;
import java.security.NoSuchAlgorithmException;
import java.security.cert.CertificateEncodingException;
import java.security.cert.X509Certificate;
import java.util.Arrays;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.List;

import javax.xml.namespace.QName;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.stream.XMLInputFactory;
import javax.xml.stream.XMLStreamReader;

import org.apache.axiom.om.OMAbstractFactory;
import org.apache.axiom.om.OMElement;
import org.apache.axiom.om.OMFactory;
import org.apache.axiom.om.OMNamespace;
import org.apache.axiom.om.impl.builder.StAXOMBuilder;
import org.apache.axiom.om.util.Base64;
import org.apache.ws.security.util.DOM2Writer;
import org.opensaml.XML;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.wso2.solutions.identity.IdentityConstants;
import org.wso2.solutions.identity.relyingparty.RelyingPartyException;
import org.wso2.solutions.identity.relyingparty.TokenVerifier;
import org.wso2.solutions.identity.relyingparty.TokenVerifierConstants;
import org.wso2.utils.ServerConfiguration;

/**
 * This is the verifier service to verify decrypted token at a relying party
 */
public class VerifierService {

    /**
     * Verify a given decrypted xmlToken
     * 
     * @param request
     *            The verification request with the token to be verified and
     *            other information
     * @return If successfully verified then an OMElement with the claims
     *         available in the xmlToken. Otherwise an OMElement with failure
     *         information.
     */
    public OMElement verify(OMElement request) throws RelyingPartyException {
        try {
            OMElement retElement = null;
            OMElement token = request
                    .getFirstChildWithName(new QName(TokenVerifierConstants.NS,
                            TokenVerifierConstants.LN_TOKEN));
            String strVal = token.getFirstChildWithName(
                    new QName(XML.SAML_NS, "Assertion")).toString();
            ByteArrayInputStream bais = new ByteArrayInputStream(strVal
                    .getBytes());

            // Convert to DOM
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            dbf.setNamespaceAware(true);
            Document doc = dbf.newDocumentBuilder().parse(bais);
            TokenVerifier tokenVerifier = new TokenVerifier();
            if (tokenVerifier.verifyDecryptedToken(doc.getDocumentElement())) {

                String issuerName = tokenVerifier.getIssuerName();

                if (IdentityConstants.SELF_ISSUED_ISSUER.equals(issuerName)) {
                    // TODO do another check to see weather it accepts self
                    // issueed
                    retElement = getSuccessResponseElement(tokenVerifier);
                } else if (this.isTrustedForUser(tokenVerifier)) {
                    retElement = getSuccessResponseElement(tokenVerifier);
                }
            }

            // has not been successfull so return failure
            if (retElement == null) {
                retElement = getFailureResponseElement(tokenVerifier);
            }
            return retElement;

        } catch (Exception e) {
            e.printStackTrace();
            throw new RelyingPartyException("verificationFailure", e);
        }

    }

    /**
     * Create the response element with the given attribute table
     * 
     * @param attrTable
     *            table of attributes to be added to the response
     * @return The response element to be returned
     */
    private OMElement getSuccessResponseElement(TokenVerifier tokenVerifer)
            throws Exception {
        Hashtable attrTable = tokenVerifer.getAttributeTable();
        OMFactory fac = OMAbstractFactory.getOMFactory();
        OMNamespace ns = fac.createOMNamespace(TokenVerifierConstants.NS,
                TokenVerifierConstants.PREFIX);
        OMElement resp = fac
                .createOMElement(TokenVerifierConstants.LN_RESP, ns);
        resp.addAttribute(TokenVerifierConstants.ATTR_STATE,
                TokenVerifierConstants.STATE_SUCCESS, null);

        Enumeration keysEnum = attrTable.keys();

        OMElement properties = fac.createOMElement(
                TokenVerifierConstants.LN_PROPERTIES, ns);

        while (keysEnum.hasMoreElements()) {
            String attrName = (String) keysEnum.nextElement();
            OMElement prop = fac.createOMElement(
                    TokenVerifierConstants.LN_PROPERTY, ns);
            prop.setText((String) attrTable.get(attrName));
            prop.addAttribute(fac.createOMAttribute(
                    TokenVerifierConstants.ATTR_NAME, null, attrName));
            properties.addChild(prop);
        }

        resp.addChild(properties);

        OMElement issuerInfoElem = fac.createOMElement(
                TokenVerifierConstants.LN_ISSUER_INFO, ns);
        List certficates = tokenVerifer.getCertificates();
        Element keyInfo = tokenVerifer.getKeyInfoElement();

        Iterator ite = certficates.iterator();
        boolean siginingSet = false;
        while (ite.hasNext()) {
            X509Certificate cert = (X509Certificate) ite.next();
            byte[] encodedCert = cert.getEncoded();
            String base64Encoded = Base64.encode(encodedCert);
            OMElement certElem = fac.createOMElement(
                    TokenVerifierConstants.LN_CERTIFICATE, ns);
            if (siginingSet == false) {
                certElem.addAttribute(TokenVerifierConstants.LN_SIGNING_CERT,
                        "true", null);
                siginingSet = true;
            }
            certElem.setText(base64Encoded);
            issuerInfoElem.addChild(certElem);
        }

        if (keyInfo != null) {
            String value = DOM2Writer.nodeToString(keyInfo);
            XMLStreamReader parser = XMLInputFactory.newInstance()
                    .createXMLStreamReader(new StringReader(value));
            StAXOMBuilder builder = new StAXOMBuilder(parser);
            OMElement omKeyInfo = builder.getDocumentElement();
            issuerInfoElem.addChild(omKeyInfo);
        }

        resp.addChild(issuerInfoElem);
        return resp;
    }

    private OMElement getFailureResponseElement(TokenVerifier tokenVerifer) {
        OMFactory fac = OMAbstractFactory.getOMFactory();
        OMNamespace ns = fac.createOMNamespace(TokenVerifierConstants.NS,
                TokenVerifierConstants.PREFIX);
        OMElement resp = fac
                .createOMElement(TokenVerifierConstants.LN_RESP, ns);
        resp.addAttribute(TokenVerifierConstants.ATTR_STATE,
                TokenVerifierConstants.STATE_FAILURE, null);

        return resp;
    }

    private boolean isTrustedForUser(TokenVerifier tokenVerifer)
            throws Exception {
        boolean isTrusted = false;

        String issuerName = tokenVerifer.getIssuerName();

        URI uri = new URI(issuerName);
        String hostName = uri.getHost();

        if (hostName == null) {
            throw new RelyingPartyException("invalidIssuerName");
        }

        ServerConfiguration serverConfig = ServerConfiguration.getInstance();
        String storeFilePath = serverConfig
                .getFirstProperty("Security.IDPKeyStore.Location");
        String passwd = serverConfig
                .getFirstProperty("Security.IDPKeyStore.Password");
        String type = serverConfig
                .getFirstProperty("Security.IDPKeyStore.Type");

        FileInputStream is = new FileInputStream(storeFilePath);
        KeyStore store = KeyStore.getInstance(type);
        store.load(is, passwd.toCharArray());

        boolean isContained = store.containsAlias(hostName);
        if (!isContained) {
            throw new RelyingPartyException("IDPNotTrusted");
        }

        X509Certificate trustedCert = (X509Certificate) store
                .getCertificate(hostName);
        List lst = tokenVerifer.getCertificates();

        if (lst.size() > 0) {
            X509Certificate signedCert = (X509Certificate) lst.get(0);
            isTrusted = compareCertificates(signedCert, trustedCert);
        } else {
            Element keyInfoElem = tokenVerifer.getKeyInfoElement();
            byte[] keyValue = trustedCert.getPublicKey().getEncoded();
            String trustedModules = Base64.encode(keyValue);

            Node node = keyInfoElem.getElementsByTagName("Modulus").item(0);
            String recievedModulus = node.getTextContent();
            if (trustedModules.equals(recievedModulus)) {
                isTrusted = true;
            }
        }

        return isTrusted;
    }

    private boolean compareCertificates(X509Certificate cert1,
            X509Certificate cert2) throws RelyingPartyException {
        boolean isEqual = true;
        byte[] th1 = getThumbPrint(cert1);
        byte[] th2 = getThumbPrint(cert2);
        if (th1 != null && th2 != null) {
            isEqual = Arrays.equals(th1, th2);
        }
        return isEqual;
    }

    private byte[] getThumbPrint(X509Certificate cert)
            throws RelyingPartyException {

        try {
            MessageDigest sha = MessageDigest.getInstance("SHA-1");
            sha.reset();
            sha.update(cert.getEncoded());

            byte[] data = sha.digest();

            return data;
        } catch (CertificateEncodingException e1) {
            throw new RelyingPartyException("encodeError", e1);

        } catch (NoSuchAlgorithmException e) {
            throw new RelyingPartyException("errorCalculatingThumbprint", e);
        }

    }
}
